"""
@author: Julian Sobott
@brief:
@description:

@external_use:

@internal_use:
"""
import unittest
from unittest import TestCase

from pynetworking.core.Data import ByteStream, pack_int


class TestByteStream(TestCase):

    def test_init(self):
        byte_string = b""
        byte_stream = ByteStream(byte_string)
        self.assertEqual(byte_stream.byte_string, b"")
        self.assertEqual(byte_stream.length, 0)
        self.assertEqual(byte_stream.idx, 0)
        self.assertEqual(byte_stream.remaining_length, 0)
        self.assertEqual(byte_stream.reached_end, True)

    def test_next_int(self):
        first_int = 102
        second_int = 2089
        example_byte_string = pack_int(first_int) + pack_int(second_int)
        byte_stream = ByteStream(example_byte_string)

        next_int = byte_stream.next_int()
        self.assertEqual(next_int, first_int)
        next_int = byte_stream.next_int()
        self.assertEqual(next_int, second_int)

    def test_next_bytes(self):
        example_byte_string = b"Hello World"
        byte_stream = ByteStream(example_byte_string)
        expected = b"Hello"
        actual = byte_stream.next_bytes(5)
        self.assertEqual(expected, actual)

    def test_next_bytes_error(self):
        example_byte_string = b"H"
        byte_stream = ByteStream(example_byte_string)
        self.assertRaises(IndexError, byte_stream.next_bytes, 5)
        self.assertRaises(AssertionError, byte_stream.next_bytes, -2)

    def test_iadd(self):
        byte_stream = ByteStream(b"Hello")
        add = b"Hello"
        byte_stream += add
        expected = b"HelloHello"
        self.assertEqual(byte_stream.byte_string, expected)
        self.assertEqual(byte_stream.reached_end, False)

    def test_remove_consumed_bytes(self):
        byte_stream = ByteStream(b"Hello World")
        byte_stream.next_bytes(len(b"Hello "))
        byte_stream.remove_consumed_bytes()
        expected = b"World"
        self.assertEqual(byte_stream.byte_string, expected)

        byte_stream = ByteStream(b"Hello World")
        byte_stream.next_bytes(len(b"Hello World"))
        byte_stream.remove_consumed_bytes()
        self.assertEqual(byte_stream.reached_end, True)


if __name__ == '__main__':
    unittest.main()



"""
@author: Julian Sobott
@brief:
@description:

@external_use:

@internal_use:

"""
import json
from functools import wraps
import time


class Ddict(dict):
    """
    Dict object where value by key and key by value is possible
    """
    def __getitem__(self, item):
        try:
            return super().__getitem__(item)
        except KeyError:
            return list(self.keys())[list(self.values()).index(item)]


class _DictEncoder(json.JSONEncoder):
    def encode(self, dict_):
        def hint_tuples(item):
            if isinstance(item, tuple):
                return {"__tuple__": True, "items": tuple(hint_tuples(e) for e in item)}
            if isinstance(item, list):
                return [hint_tuples(e) for e in item]
            if isinstance(item, dict):
                return {key: hint_tuples(value) for key, value in item.items()}
            else:
                return item

        return super().encode(hint_tuples(dict_))


def _hinted_tuple_hook(obj):
    if "__tuple__" in obj:
        return tuple(obj["items"])
    else:
        return obj


def dump_dict_to_json(dict_):
    enc = _DictEncoder()
    return enc.encode(dict_)


def load_dict_from_json(json_):
    return json.loads(json_, object_hook=_hinted_tuple_hook)


def time_func(f):
    @wraps(f)
    def wrap(*args, **kw):
        ts = time.time()
        result = f(*args, **kw)
        te = time.time()
        print(f'func:{f.__name__} took: {te - ts} sec')
        return result
    return wrap



"""
:module: pynetworking.Communication_general
:synopsis: Important functions, classes for communication. Main module of pynetworking.
:author: Julian Sobott

public classes
----------------
Communicator, Connector, SocketAddress, Functions, to_server_id

.. autoclass:: Communicator
    :members:
    :undoc-members:
    :private-members:

.. autoclass:: Connector
    :members:
    :undoc-members:

.. autoclass:: SingleConnector
    :members:
    :undoc-members:

.. autoclass:: MultiConnector
    :members:
    :undoc-members:

.. autoclass:: MultiConnector
    :members:
    :undoc-members:

.. autoclass:: Functions
    :noindex:
    :members:
    :undoc-members:


public functions
----------------

.. autofunction:: to_server_id
.. autofunction:: to_client_id
.. autofunction:: set_encrypted_communication

private classes
-----------------

.. autoclass:: PacketBuilder
    :members:
    :undoc-members:

.. autoclass:: MetaFunctionCommunicator
    :members: __call__, __getattribute__, __getattr__

.. autoclass:: MetaSingletonConnector
    :members:
    :undoc-members:

.. autoclass:: FunctionExecutionThread

"""
import threading
import socket
import time
import traceback
import sys
from typing import Tuple, List, Dict, Optional, Callable, Any, Type, Union

from pynetworking.core.Cryptography import Cryptographer
from pynetworking.core.Logging import logger
from pynetworking.core.Packets import Packet, DataPacket, FunctionPacket, FileMetaPacket, Header
from pynetworking.core.ID_management import IDManager, remove_manager
from pynetworking.core.Data import ByteStream, File

SocketAddress = Tuple[str, int]

"""Server and client ids are different to separate them, when server and client both run on the same machine."""
CLIENT_ID_START = 30
SERVER_ID_START = 0  # Max 30 servers

ENCRYPTED_COMMUNICATION = True


def set_encrypted_communication(value: bool):
    """Allows to deactivate encrypted communication."""
    global ENCRYPTED_COMMUNICATION
    ENCRYPTED_COMMUNICATION = value


def to_client_id(id_: int) -> int:
    return int(id_ + CLIENT_ID_START)


def to_server_id(id_: int) -> int:
    return int(id_ + SERVER_ID_START)


class Communicator(threading.Thread):
    """Super class for all communicators. It handles a tcp-socket connection. Can send and receive packets and
    handles them. Responsible for connecting to another tcp-socket.
    """
    CHUNK_SIZE = 4096

    def __init__(self, address: SocketAddress, id_, socket_connection: socket.socket = None, from_accept=False,
                 on_close: Optional[Callable[['Communicator'], Any]] = None, local_functions=Type['Functions']) -> None:
        super().__init__(name=f"{'Client' if from_accept else 'Server'}_Communicator_thread_{id_}")
        self._recv_timeout = 1

        self._socket_connection = socket_connection
        if socket_connection is not None:
            self._socket_connection.settimeout(self._recv_timeout)

        self._address = address
        self._id = id_
        self._is_connected = from_accept
        self._keep_connection = not from_accept
        self._packets: List[Packet] = []
        self._exit = threading.Event()
        self._time_till_next_check = 0.3
        self._on_close = on_close
        self._functions: Type['Functions'] = local_functions
        self._auto_execute_functions = from_accept
        self._closed = False
        self.wait_for_response_timeout = float("inf")
        self.cryptographer = Cryptographer()

    def run(self) -> None:
        """Connects to a tcp-socket. When it is connected, it listens to packets, that are sent from the other
        socket."""
        if not self._is_connected:
            self._connect()
        if self.is_connected():
            self._wait_for_new_input()

    def send_packet(self, packet: Packet) -> bool:
        """Set the proper ids and converts/packs the packet into bytes. Sends the bytes string."""
        IDManager(self._id).set_ids_of_packet(packet)
        send_data = packet.pack()
        successfully_sent = self._send_bytes(send_data)
        if not successfully_sent:
            logger.error("Could not send packet: %s", str(packet))
        return successfully_sent

    def wait_for_response(self):
        """Waits till a data-packet is received and returns it. If a function packet is received instead it is
        executed first."""
        waited = 0.
        while not self._exit.is_set():
            next_global_id = IDManager(self._id).get_next_outer_id()
            try:
                next_packet = self._packets.pop(0)
                actual_outer_id = next_packet.header.id_container.global_id
                if actual_outer_id > next_global_id:
                    logger.error(f"Packet lost! Expected outer_id: {next_global_id}. Got instead: {actual_outer_id}")
                    # TODO: handle
                elif actual_outer_id < next_global_id:
                    logger.error(f"Unhandled Packet! Expected outer_id: {next_global_id}. "
                                 f"actual: {actual_outer_id}, Communicator id: {self._id}")
                    # TODO: handle (if possible)
                else:
                    if isinstance(next_packet, Packet):
                        self._handle_packet(next_packet)
                    if isinstance(next_packet, FunctionPacket):
                        """execute and keep waiting for data"""
                    elif isinstance(next_packet, DataPacket):
                        return next_packet
                    elif isinstance(next_packet, FileMetaPacket):
                        """File is already transmitted."""
                        return DataPacket(**{"return": File.from_meta_packet(next_packet)})
                    else:
                        logger.error(f"Received not implemented Packet class: {type(next_packet)}")

            except IndexError:
                pass  # List is empty -> wait
            self._exit.wait(self._time_till_next_check)
            waited += self._time_till_next_check
            if waited > self.wait_for_response_timeout >= 0:
                logger.info("wait_for_response waited too long")
                raise TimeoutError("wait_for_response waited too long")

    def _connect(self, seconds_till_next_try: float = 2, timeout: float = -1) -> bool:
        waited = 0
        while not self._exit.is_set() and not self._is_connected:
            # TODO: Get timeout time from Connector.connect
            try:
                self._socket_connection = socket.create_connection(self._address)
                self._socket_connection.settimeout(self._recv_timeout)
                self._is_connected = True
                logger.info(f"Successfully connected to: {str(self._address)}")
                return True
            except ConnectionRefusedError:
                logger.info("Could not connect to server with address: (%s)", str(self._address))
            except socket.gaierror:
                raise ValueError(
                    f"Address error. {self._address} is not a valid address. Address must be of type {SocketAddress}")
            except OSError as e:
                logger.error("Is already connected to server")
                logger.debug(e)
                self._is_connected = True

            self._exit.wait(seconds_till_next_try)
            waited += seconds_till_next_try
            if waited > timeout >= 0:
                logger.warning("Connection timeout")
                return False
        return False

    def _wait_for_new_input(self):
        """Loop that is constantly receiving or waiting for new packets from the tcp-connection."""
        plain_byte_stream = ByteStream(b'')
        encrypted_byte_stream = ByteStream(b'')
        while not self._exit.is_set():
            if not self._exit.is_set() and not self._is_connected:
                if self._keep_connection:
                    self._connect()
                else:
                    self.stop(is_same_thread=True)
                    continue
            packet = self._recv_packet(plain_byte_stream, encrypted_byte_stream)
            if packet is not None:
                if isinstance(packet, FileMetaPacket):
                    self._recv_file(packet, plain_byte_stream, encrypted_byte_stream)
                    self._packets.append(packet)
                elif self._auto_execute_functions and isinstance(packet, FunctionPacket):
                    func_thread = FunctionExecutionThread(self._id, packet, self._handle_packet)
                    func_thread.start()
                else:
                    self._packets.append(packet)

    def _recv_data(self, plain_byte_stream: ByteStream, encrypted_byte_stream: ByteStream) -> \
            Optional[bytes]:
        """Returns every chunk, that can be decrypted"""
        data = plain_byte_stream.next_all_bytes()
        plain_byte_stream.remove_consumed_bytes()
        if len(data) > 0:
            return data
        encrypted_data = encrypted_byte_stream.next_all_bytes()
        encrypted_byte_stream.remove_consumed_bytes()
        try:
            while not self._exit.is_set():
                chunk_data = self._socket_connection.recv(self.CHUNK_SIZE)
                if chunk_data == b"":
                    logger.info("Connection reset, (%s)", str(self._address))
                    self._is_connected = False
                    return None
                else:
                    if not self.cryptographer.is_encrypted_communication:
                        return chunk_data
                    if b"%%" in chunk_data:
                        corresponding_data = chunk_data.split(b"%%")[0]
                        encrypted_data += corresponding_data
                        encrypted_byte_stream += chunk_data[len(corresponding_data) + 2:]
                        data += self.cryptographer.decrypt(encrypted_data)
                        return data
                    else:
                        encrypted_data += chunk_data

        except ConnectionResetError:
            if not self._exit.is_set():
                logger.warning(f"Connection reset at ID({self._id}), {self._address}")
            self._is_connected = False

        except ConnectionAbortedError:
            if not self._exit.is_set():
                logger.warning(f"Connection aborted at ID({self._id}), {self._address}")
            self._is_connected = False

        except OSError as e:
            if not isinstance(e, socket.timeout):
                if not self._exit.is_set():
                    logger.warning("TCP connection closed while listening")
                self._is_connected = False

    def _recv_packet(self, plain_byte_stream: ByteStream, encrypted_byte_stream: ByteStream,) -> Optional[Packet]:
        """Receives bytes till a packet can be build. This packet is returned"""
        packet_builder = PacketBuilder(plain_byte_stream)
        while not self._exit.is_set():
            chunk_data = self._recv_data(plain_byte_stream, encrypted_byte_stream)
            if chunk_data == b"" or chunk_data is None:
                return None
            else:
                possible_packet = packet_builder.add_chunk(chunk_data)
                logger.debug(possible_packet)
                if possible_packet is not None:
                    return possible_packet

    def _recv_file(self, file_meta_packet: FileMetaPacket, plain_byte_stream: ByteStream,
                   encrypted_byte_stream: ByteStream,) -> None:
        """Receives bytes, till the file is fully received. The file is saved at the destination, given in the
        file_meta_packet."""
        file_path = file_meta_packet.dst_path
        existing_bytes = plain_byte_stream.next_all_bytes()
        with open(file_path, "wb+") as file:
            file.write(existing_bytes)
        remaining_bytes = file_meta_packet.file_size - len(existing_bytes)
        appended_data = b""
        with open(file_path, "ab") as file:
            while remaining_bytes > 0:
                data = self._recv_data(plain_byte_stream, encrypted_byte_stream)
                if data is None:
                    logger.error("Connection aborted, while receiving file!")
                    return
                write_data = data[:remaining_bytes]
                appended_data = data[remaining_bytes:]
                file.write(write_data)
                remaining_bytes -= len(write_data)
            plain_byte_stream += appended_data

    def _send_bytes(self, byte_string: bytes) -> bool:
        if not self._is_connected:
            self._connect(timeout=2)
        try:
            encrypted_message = self.cryptographer.encrypt(byte_string)
            if self.cryptographer.is_encrypted_communication:
                send_message = encrypted_message + b"%%"
            else:
                send_message = encrypted_message
            sent = self._socket_connection.sendall(send_message)
            # returns None on success
            return sent is None
        except OSError:
            logger.error(f"Could not send bytes {byte_string}")
            return False

    def _handle_packet(self, packet):
        """Adjusts all ids in the IDManager. Handles function-packets."""
        IDManager(self._id).update_ids_by_packet(packet)
        if isinstance(packet, FunctionPacket):
            self._received_function_packet(packet)

    def _received_function_packet(self, packet: FunctionPacket) -> None:
        """Executes the function, with all args. Packs the return value or the exception in a data-packet and sends
        it back. If a pynetworking.File is returned, a FileMetaPacket + the file itself is sent."""
        func = packet.function_name
        args = packet.args
        kwargs = packet.kwargs
        try:
            ret_value = self._functions.__getattr__(func)(*args, **kwargs)
            if isinstance(ret_value, File):
                return self._send_file(ret_value)
        except:
            ret_value = ExceptionObject(*sys.exc_info())

        ret_kwargs = {"return": ret_value}
        data_packet = DataPacket(**ret_kwargs)
        self.send_packet(data_packet)

    def _send_file(self, file: File):
        """Creates a FileMetaPacket, that is sent and followed by the file_content."""
        file_meta_packet = FileMetaPacket(file.src_path, file.size, file.dst_path)
        self.send_packet(file_meta_packet)
        with open(file.src_path, "rb") as f:
            file_data = f.read(self.CHUNK_SIZE)
            while len(file_data) > 0:
                self._send_bytes(file_data)
                file_data = f.read(self.CHUNK_SIZE)

    def stop(self, is_same_thread=False) -> None:
        """Stops all listening and the thread is joined. Send processes are not stopped and it the thread first
        stops when all data is sent."""
        if self._closed:
            logger.debug("Prevented closing already closed communicator")
        else:
            communicator_side = "Client" if self._id >= CLIENT_ID_START else "Server"
            logger.info(f"Stopping {communicator_side}side communicator: {self._id}")
            self._exit.set()
            if self._socket_connection is not None:
                self._socket_connection.close()
            self._is_connected = False
            if not is_same_thread:
                self.join()
            remove_manager(self._id)
            try:
                self._on_close(self)
            except TypeError:
                pass  # no function provided
            self._closed = True

    def is_connected(self) -> bool:
        return self._is_connected

    def get_id(self) -> int:
        return self._id


class PacketBuilder:
    """Builds a packet from bytes chunk data. Bytes must be added till a packet can be built. This packet is returned"""

    def __init__(self, byte_stream: ByteStream) -> None:
        self.byte_stream = byte_stream
        self.current_header: Optional[Header] = None

    def add_chunk(self, byte_string: bytes) -> Optional[Packet]:
        """Byte string is from receiving bytes from the tcp-connection. This function rebuilds the packet from it,
        if there was enough data added."""
        self.byte_stream += byte_string
        if self.current_header is None and self.byte_stream.length >= Header.LENGTH_BYTES:
            self.current_header = Header.from_bytes(self.byte_stream)
        if self.current_header and self.byte_stream.remaining_length >= self.current_header.specific_data_size:
            packet = Packet.from_bytes(self.current_header, self.byte_stream)
            self.byte_stream.remove_consumed_bytes()
            self.current_header = None
            return packet
        return None


class MetaFunctionCommunicator(type):
    """
    Magic class.

    This class makes it possible, that functions, that are located at the other side, and should be
    transmitted of the network can be called, like they were local functions. The dunder functions of python play an
    important role in this class. Python can intercept the process of getting an attribute with `__getattribute__`.
    This function every time something is called with `x.foo` or `x.bar()`. Normally this just returns the attribute,
    but in this case we treat every attribute (except special attributes) as a function call. This function call is
    then packed into a function-packet with all its args and kwargs and is sent over the tcp-connection to the other
    side. The data-packet, that is returned, is unpacked and returned to the caller. If an exception was risen at the
    other side it also raises locally.

    This class is the metaclass for following Function classes."""

    def __call__(cls, *args, **kwargs):
        """A subclass can be called (not necessary). It is possible to call it with the `timeout` argument
        (e.g. `remote_functions(timeout=2.5).function(...)`). This way it is waited till the data-packet arrives or
        the timeout raises."""
        try:
            timeout = kwargs["timeout"]
            connector: Connector = cls.__getattr__("_connector")
            if connector and connector.communicator:
                connector.communicator.wait_for_response_timeout = timeout
        except KeyError:
            pass
        return cls

    def __getattribute__(self, item: str):
        """Intercepts the calling process of functions. To allow calling special variables there is this if block at
        the beginning.

        NOTE: that if you want to access any other attribute, you need to call it with the dunder functions, that are
        defined in the if block."""
        if item == "__getattr__":
            return type.__getattribute__(self, item)
        if item == "__setattr__":
            return type.__setattr__
        if item == "__call__":
            return type.__call__
        if item.startswith("__", 0, 2):
            return type.__getattribute__(self, item)

        def container(*args, **kwargs) -> Any:
            """Container works like a decorator container. The 'decoration' is, that it sends a function-packet and
            receives and unpacks a data-packet."""
            function_name = item
            # send function packet
            connector: Connector = self.__getattr__("_connector")
            function_packet = FunctionPacket(function_name, *args, **kwargs)
            if connector is None or connector.communicator is None:
                raise ConnectionError(
                    "Communicator is not connected!"
                    "Connect first to a server with `ServerCommunicator.connect(server_address)´")
            sent_packet = connector.communicator.send_packet(function_packet)
            if not sent_packet:
                raise ConnectionError("Could not send function to server. Check connection to server.")

            try:
                data_packet = connector.communicator.wait_for_response()
            except TimeoutError as e:
                raise e
            # unpack data packet
            return_values = data_packet.data["return"]
            if isinstance(return_values, ExceptionObject):
                # An exception was thrown at the other side!
                raise return_values.exec_type(return_values.get_formatted())
            return return_values

        return container

    def __getattr__(self, attr_name: str) -> Any:
        """Allows accessing attributes without interception. Usage: <class_object>.__getattr__("<attribute_name>")"""
        attribute = type.__getattribute__(self, attr_name)
        return attribute


class MetaSingletonConnector(type):
    """Allows singleton like connectors. Each connector is identified, by its id."""
    _instances: Dict[int, 'Connector'] = {}

    def __call__(cls, *args, **kwargs) -> 'Connector':
        id_: int = args[0]
        if id_ not in cls._instances:
            cls._instances[id_] = super(MetaSingletonConnector, cls).__call__(*args, **kwargs)
        return cls._instances[id_]

    @classmethod
    def remove(mcs, id_: int) -> 'Connector':
        return mcs._instances.pop(id_)

    @classmethod
    def remove_all(mcs) -> Dict[int, 'Connector']:
        ret = dict(mcs._instances)
        mcs._instances = {}
        return ret


class Connector:
    """Super class for :class:`MultiConnector` and :class:`SingleConnector`. The subclasses are responsible for the
    connection and communication.

    :ivar remote_functions: Class with all functions, that are available at the remote side.
    :ivar local_functions: Class with all functions, that are locally available.
    :ivar communicator: instance of :class:`Communicator`.
    :ivar _id:
    """
    remote_functions: Optional[Type['Functions']] = None
    local_functions: Optional[Type['Functions']] = None

    communicator: Optional[Communicator] = None
    _id = to_client_id(0)
    _exchanged_keys = False

    @staticmethod
    def connect(connector: Union['Connector', Type['SingleConnector']], addr: SocketAddress, blocking=True,
                timeout=float("inf"), exchange_keys_function=None) -> Optional[bool]:
        """Connects the passed `connector` to the server. This is done, by creating a new :class:`Communicator`,
        that connects to the server. Also creates a relation between this connector and the remote_functions. The
        connection process can be executed in a separate thread."""
        if connector.communicator is None or not connector.communicator.is_connected():
            connector.communicator = Communicator(addr, id_=connector._id, local_functions=connector.local_functions)
            try:
                connector.remote_functions.__setattr__(connector.remote_functions, "_connector", connector)
                # self argument, must be provided
            except TypeError:
                raise AttributeError("Communicator functions are not set.")

            connector.communicator.start()

            def try_connect():
                waited = 0.
                wait_time = 0.01
                while not connector.communicator.is_connected() and waited < timeout:
                    time.sleep(wait_time)
                    waited += wait_time
                if waited >= timeout:
                    logger.info(f"Stopped trying to connect to server after {int(waited)} seconds due to timeout")
                    connector.communicator.stop()
                    connector.communicator = None
                    return False
                    # raise TimeoutError(f"Stopped trying to connect to server after {int(waited)} seconds")
                elif ENCRYPTED_COMMUNICATION is True:
                    exchange_keys_function(connector)
                return True

            if blocking:
                return try_connect()
            else:
                threading.Thread(target=try_connect, name=f"Connector_{connector._id}").start()

    @staticmethod
    def close_connection(connector: Union['Connector', Type['SingleConnector']], blocking=True,
                         timeout=float("inf")) -> None:
        """Closes the connection to the server. If blocking is True, it will wait till the communicator stopped the
        connection."""
        if connector.communicator is not None:
            connector.communicator.stop()
            if blocking:
                waited = 0.
                wait_time = 0.01
                while connector.communicator and connector.communicator.is_connected() and waited < timeout:
                    time.sleep(wait_time)
                    waited += wait_time
            connector.communicator = None

    @staticmethod
    def is_connected(connector: Union['Connector', Type['SingleConnector']]) -> bool:
        if connector.communicator is None:
            return False
        return connector.communicator.is_connected()

    @staticmethod
    def get_id(connector: Union['Connector', Type['SingleConnector']]):
        return connector._id


class MultiConnector(Connector, metaclass=MetaSingletonConnector):
    """Connector that allows creating multiple instances. Every instance is handled like a Singleton."""

    def __init__(self, id_: int) -> None:
        self._id = to_client_id(id_)
        self.communicator: Optional[Communicator] = None
        self._exchanged_keys = False

    def connect(self: Connector, addr: SocketAddress, blocking=True, timeout=float("inf"), exchange_keys_function=None) \
            -> bool:
        return super().connect(self, addr, blocking, timeout, exchange_keys_function)

    def close_connection(self: Connector, blocking=True, timeout=float("inf")) -> None:
        return super().close_connection(self, blocking, timeout)

    @staticmethod
    def close_all_connections() -> None:
        all_instances = MetaSingletonConnector.remove_all()
        for id_, connector in all_instances.items():
            connector.close_connection(connector)

    def is_connected(self) -> bool:
        return super().is_connected(self)

    def get_id(self):
        return super().get_id(self)

    @property
    def exchanged_keys(self):
        return self._exchanged_keys


class SingleConnector(Connector):
    """Only static accessible. Therefore only a single connector (per address) per machine possible. For most
    applications this class is sufficient and the :class:`MultiConnector` isn't needed."""

    @classmethod
    def connect(cls, addr: SocketAddress, blocking=True, timeout=float("inf"), exchange_keys_function=None) \
            -> Optional[bool]:
        return super().connect(cls, addr, blocking, timeout, exchange_keys_function)

    @classmethod
    def close_connection(cls, blocking=True, timeout=float("inf")) -> None:
        return super().close_connection(cls, blocking, timeout)

    @classmethod
    def is_connected(cls) -> bool:
        return super().is_connected(cls)

    @classmethod
    def get_id(cls):
        return super().get_id(cls)


class Functions(metaclass=MetaFunctionCommunicator):
    """Static class that contains all available local and remote functions. All functions must be stored in the
        :attr:`__dict__` attribute.

            """
    _connector: Optional[Connector] = None


class FunctionExecutionThread(threading.Thread):
    """Executes a received function in a separate thread."""
    def __init__(self, id_: int, function_packet: FunctionPacket, handle_packet: Callable) -> None:
        super().__init__(name=f"FunctionExecutionThread_{id_}")
        self._id = id_
        self._function_packet = function_packet
        self._handle_packet = handle_packet

    def run(self):
        self._handle_packet(self._function_packet)

    @property
    def id(self):
        return self._id


class ExceptionObject:

    def __init__(self, exc_type, exc_value, exc_traceback):
        self._lines = traceback.format_exception(exc_type, exc_value, exc_traceback)
        self.tb = traceback.extract_tb(exc_traceback)
        self.exec_type = exc_type

    def print_exception(self):
        for line in self._lines:
            print(line, end="", file=sys.stderr)

    def get_formatted(self):
        return "An Error occurred on the other side!\n"+"".join(self._lines)



"""
:synopsis: A simple echo server
:author: Julian Sobott

The server always prints the received message from the client.
If the client writes 'bye', the client calls a function at the client and prints the result.

"""

import pynetworking as net

# define the address at which the server listen and the clients connect to
server_address = "127.0.0.1", 5000

# Only log errors
net.Logging.logger.setLevel(40)


def main():
    # open and start the server and close afterwards
    # As arguments, the address, to which the server listens to and the class that overrides the ClientCommunicator is
    # needed
    with net.ClientManager(server_address, ClientCommunicator):
        # Connect a client to the server
        # The connection is stored statically. This means no object needs to be created and passed to other functions.
        # Once the connection is established, you can call the functions from everywhere
        ServerCommunicator.connect(server_address, blocking=True, timeout=2)

        # create an alias for all server functions (Not necessary, but code looks better)
        server = ServerCommunicator.remote_functions

        while True:
            message = input(">>>")
            server.recv_message_from_client(message)
            if message == "bye":
                break

        # Close the connection to the server at client-side
        ServerCommunicator.close_connection()


class ClientFunctions(net.ClientFunctions):
    # A class that defines every method that can be called at a client from the server
    @staticmethod
    def get_last_words() -> str:
        return "Please don't leave me alone!"


class ServerFunctions(net.ServerFunctions):
    # A class that defines every method that can be called at the server from a client
    @staticmethod
    def recv_message_from_client(message: str) -> None:
        print(f"Client: {message}")
        if message == "bye":
            client: ClientCommunicator = net.ClientManager().get()
            clients_last_words = client.remote_functions.get_last_words()
            print(clients_last_words)


class ClientCommunicator(net.ClientCommunicator):
    # Server-side class that sets the available functions and make them available for communication
    local_functions = ServerFunctions
    remote_functions = ClientFunctions


class ServerCommunicator(net.ServerCommunicator):
    # Client-side class that sets the available functions and make them available for communication
    local_functions = ClientFunctions
    remote_functions = ServerFunctions


if __name__ == '__main__':
    main()



"""
@author: Julian Sobott
@brief:
@description:

@external_use:

@internal_use:
"""
import pynetworking as net

from pynetworking.core.Logging import logger


class _DummyServerFunctions(net.ServerFunctions):

    @staticmethod
    def return_client_id() -> int:
        return net.ClientManager().get().id

    @staticmethod
    def no_arg_no_ret() -> None:
        return no_arg_no_ret()

    @staticmethod
    def no_arg_ret() -> bool:
        return no_arg_ret()

    @staticmethod
    def incrementer(number: int) -> int:
        return incrementer(number)

    @staticmethod
    def immutable_args_ret(name: str, age: int, children: tuple) -> str:
        return immutable_args_ret(name, age, children)

    @staticmethod
    def args_ret_object(name, age) -> 'DummyPerson':
        return args_ret_object(name, age)

    @staticmethod
    def class_args_ret(person: 'DummyPerson') -> tuple:
        return class_args_ret(person)

    @staticmethod
    def huge_args_huge_ret(*args):
        return huge_args_huge_ret(*args)

    @staticmethod
    def func_in_func(start: int) -> bool:
        return func_in_func(start)

    @staticmethod
    def many_func_in_func():
        return many_func_in_func()

    @staticmethod
    def server_faculty(number: int) -> int:
        return server_faculty(number)

    @staticmethod
    def client_faculty(number: int) -> int:
        return client_faculty(number)

    @staticmethod
    def get_file(file_path: str, destination_path: str) -> net.File:
        return get_file(file_path, destination_path)


class _DummyClientFunctions(net.ServerFunctions):
    @staticmethod
    def no_arg_no_ret() -> None:
        return no_arg_no_ret()

    @staticmethod
    def no_arg_ret() -> bool:
        return no_arg_ret()

    @staticmethod
    def incrementer(number: int) -> int:
        return incrementer(number)

    @staticmethod
    def immutable_args_ret(name: str, age: int, children: tuple) -> str:
        return immutable_args_ret(name, age, children)

    @staticmethod
    def args_ret_object(name, age) -> 'DummyPerson':
        return args_ret_object(name, age)

    @staticmethod
    def class_args_ret(person: 'DummyPerson') -> tuple:
        return class_args_ret(person)

    @staticmethod
    def huge_args_huge_ret(*args):
        return huge_args_huge_ret(*args)

    @staticmethod
    def func_in_func(start: int) -> bool:
        return func_in_func(start)

    @staticmethod
    def many_func_in_func():
        return many_func_in_func()

    @staticmethod
    def server_faculty(number: int) -> int:
        return server_faculty(number)

    @staticmethod
    def client_faculty(number: int) -> int:
        return client_faculty(number)


class DummyServerCommunicator(net.ServerCommunicator):
    remote_functions = _DummyServerFunctions
    local_functions = _DummyClientFunctions


class DummyMultiServerCommunicator(net.MultiServerCommunicator):
    remote_functions = _DummyServerFunctions
    local_functions = _DummyClientFunctions


class DummyClientCommunicator(net.ClientCommunicator):
    remote_functions = _DummyClientFunctions
    local_functions = _DummyServerFunctions


class DummyPerson:
    def __init__(self, name, age):
        self.name = name
        self.age = age

    def __eq__(self, other):
        if isinstance(other, DummyPerson):
            return (self.name == other.name
                    and self.age == other.age)


def no_arg_no_ret() -> None:
    print("no_arg_no_ret() called")


def no_arg_ret() -> bool:
    print("no_arg_ret() called")
    return True


def incrementer(number: int) -> int:
    return number + 1


def immutable_args_ret(name: str, age: int, children: tuple) -> str:
    return f"{name} is {age} old and has {len(children)}: {children}"


def args_ret_object(name, age) -> DummyPerson:
    return DummyPerson(name, age)


def class_args_ret(person: DummyPerson) -> tuple:
    name = person.name
    age = person.age
    return name, age


def huge_args_huge_ret(*args):
    return args


def func_in_func(start: int) -> bool:
    ret = net.ClientManager().get().remote_functions.incrementer(start)
    return ret + 1


def many_func_in_func():
    return net.ClientManager().get().remote_functions.client_faculty(5)


def server_faculty(number: int) -> int:
    if number <= 1:
        return number
    return number * net.ClientManager().get().remote_functions.client_faculty(number - 1)


def client_faculty(number: int) -> int:
    if number <= 1:
        return number
    return number * DummyServerCommunicator.remote_functions.server_faculty(number - 1)


def get_file(file_path: str, destination_path: str) -> net.File:
    return net.File(file_path, destination_path)




"""
:module: pynetworking.Cryptography
:synopsis: Access to Cryptography
:author: Julian Sobott

public classes
---------------

.. autoclass:: Cryptographer
   :members:
   :undoc-members:
"""
from cryptography.fernet import Fernet
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.asymmetric import padding
from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.backends import default_backend as crypto_default_backend


class Cryptographer:
    """Functions for encryption. Supports symmetric and asymmetric keys.

    :ivar is_encrypted_communication: Signals whether the messages are plain or encrypted.
    :ivar _communication_key: All messages are encrypted with this key.
    :ivar _public_key: Needed to exchange the communication key.
    :ivar _private_key: Needed to exchange the communication key.
    :ivar _fernet: Decrypts and encrypts messages.
    """
    def __init__(self):
        self._communication_key = None
        self._public_key = None
        self._private_key = None
        self._fernet = None
        self.is_encrypted_communication = False

    @property
    def communication_key(self):
        return self._communication_key

    @communication_key.setter
    def communication_key(self, key):
        """Sets the communication key, is_encrypted_communication and fernet"""
        if self.communication_key is None:
            self.is_encrypted_communication = True
            self._fernet = Fernet(key)
            self._communication_key = key

    def encrypt(self, byte_string: bytes) -> bytes:
        if not self.is_encrypted_communication:
            return byte_string
        return self._fernet.encrypt(byte_string)

    def decrypt(self, byte_string: bytes) -> bytes:
        if not self.is_encrypted_communication:
            return byte_string
        return self._fernet.decrypt(byte_string)

    def generate_key_pair(self) -> tuple:
        self._private_key = rsa.generate_private_key(
            backend=crypto_default_backend(),
            public_exponent=65537,
            key_size=2048
        )
        self._public_key = self._private_key.public_key()
        return self._private_key, self._public_key

    @staticmethod
    def generate_communication_key():
        """Generates a symmetric key"""
        return Fernet.generate_key()

    def get_serialized_public_key(self) -> bytes:
        return self._public_key.public_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PublicFormat.SubjectPublicKeyInfo
        )

    def public_key_from_serialized_key(self, serialized_key: bytes):
        """Set the public key, by deserialize the serialized public key."""
        self._public_key = serialization.load_pem_public_key(
            serialized_key,
            backend=crypto_default_backend()
        )
        return self._public_key

    def encrypt_with_public_key(self, message: bytes) -> bytes:
        return self._public_key.encrypt(
            message,
            padding.OAEP(
                mgf=padding.MGF1(algorithm=hashes.SHA256()),
                algorithm=hashes.SHA256(),
                label=None
            )
        )

    def decrypt_with_private_key(self, cipher_text: bytes) -> bytes:
        return self._private_key.decrypt(
            cipher_text,
            padding.OAEP(
                mgf=padding.MGF1(algorithm=hashes.SHA256()),
                algorithm=hashes.SHA256(),
                label=None
            )
        )



"""
@author: Julian Sobott
@created: 13.11.2018
@brief:
@description:

@external_use:

@internal_use:

"""
import logging


logger = logging.getLogger(__name__)
handler = logging.StreamHandler()
formatter = logging.Formatter(
        '%(levelname)-8s %(message)s \t\t(%(filename)s %(lineno)d) \t%(relativeCreated).9s')
handler.setFormatter(formatter)
logger.addHandler(handler)
logger.setLevel(logging.DEBUG)





"""
High level network communication

This tool abstracts network communication to a level, where the end user don`t has to care about
network communication. Server side functions can be called at the client as they were local and vise versa.
Functions may be called with parameters and may return values.

"""
from common import *



"""
:synopsis: Necessary setup for the pynetworking library
"""
import pynetworking as net


class ServerFunctions(net.ServerFunctions):
    from Server import greet_client, server_faculty


class ClientFunctions(net.ClientFunctions):
    from Client import client_faculty, client_func


class ServerCommunicator(net.ServerCommunicator):
    remote_functions = ServerFunctions
    local_functions = ClientFunctions


class ClientCommunicator(net.ClientCommunicator):
    remote_functions = ClientFunctions
    local_functions = ServerFunctions


#: alias for all server functions
server = ServerCommunicator.remote_functions


def get_client() -> ClientCommunicator:
    """Main purpose is to supply better auto completions"""
    return net.ClientManager().get()




"""
:synopsis: All functions located at the client_side
"""
import pynetworking as net
import Interface


def client_func(name) -> bool:
    print(f"Hello {name}")
    return True


def client_faculty(number: int) -> int:
    if number <= 1:
        return number
    return number * Interface.ServerCommunicator.remote_functions.server_faculty(number - 1)


if __name__ == '__main__':
    address = ("127.0.0.1", 5000)
    with net.ClientManager(address, Interface.ClientCommunicator):
        Interface.ServerCommunicator.connect(address)
        ret_value = Interface.ServerCommunicator.remote_functions.server_faculty(5)
        print(f"Client: {ret_value}")

        ret_value = Interface.ServerCommunicator.remote_functions.greet_client("Walter")
        print(f"Client: {ret_value}")
        Interface.ServerCommunicator.close_connection()




from pynetworking.core.Communication_client import ServerCommunicator, ServerFunctions, MultiServerCommunicator
from pynetworking.core.Communication_server import ClientCommunicator, ClientFunctions, ClientManager
from pynetworking.core.Data import File
import pynetworking.core.utils
import pynetworking.core.Logging as Logging
from pynetworking.core import global_data

Logging.logger.setLevel(40)



"""
@author: Julian Sobott
@brief:
@description:

@external_use:

@internal_use:

@TODO: test id's when packets are implemented
"""
from unittest import TestCase

from pynetworking.core.ID_management import *
from pynetworking.core.Packets import FunctionPacket, DataPacket, Packet, Header
from pynetworking.core.Data import ByteStream
from pynetworking.core.Logging import logger


class TestIDManager(TestCase):

    def doCleanups(self):
        remove_manager(0)
        remove_manager(1)
        remove_manager(2)

    def test_multiple_same_managers(self):
        manager1_0 = IDManager(1)
        manager1_1 = IDManager(1)
        self.assertEqual(manager1_0, manager1_1)

    def test_multiple_different_managers(self):
        manager1 = IDManager(1)
        manager2 = IDManager(2)
        self.assertNotEqual(manager1, manager2)

    def test_remove_manager(self):
        manager1_0 = IDManager(1)
        remove_manager(1)
        manager1_1 = IDManager(1)
        self.assertNotEqual(manager1_0, manager1_1)

    def test_one_func_packet(self):
        packet = FunctionPacket("dummy")
        IDManager(0).set_ids_of_packet(packet)
        expected_ids = (0, 0)
        self.assertEqual(expected_ids, packet.header.id_container.get_ids())
        expected_function_stack = [0]
        self.assertEqual(expected_function_stack, IDManager(0).get_function_stack())

    def test_func_data(self):
        func_packet = FunctionPacket("dummy")
        IDManager(0).set_ids_of_packet(func_packet)

        data_packet = DataPacket(ret="Nothing")
        IDManager(0).set_ids_of_packet(data_packet)
        expected_ids = (0, 1)
        self.assertEqual(expected_ids, data_packet.header.id_container.get_ids())
        expected_function_stack = []
        self.assertEqual(expected_function_stack, IDManager(0).get_function_stack())

    def test_func_func_data_data(self):
        self.assertEqual([], IDManager(0).get_function_stack())

        func_packet_0 = FunctionPacket("dummy")
        IDManager(0).set_ids_of_packet(func_packet_0)
        self.assertEqual((0, 0), func_packet_0.header.id_container.get_ids())
        self.assertEqual([0], IDManager(0).get_function_stack())

        func_packet_1 = FunctionPacket("Dummy")
        IDManager(0).set_ids_of_packet(func_packet_1)
        self.assertEqual((1, 1), func_packet_1.header.id_container.get_ids())
        self.assertEqual([0, 1], IDManager(0).get_function_stack())

        data_packet_1 = DataPacket(ret="Nothing")
        IDManager(0).set_ids_of_packet(data_packet_1)
        self.assertEqual((1, 2), data_packet_1.header.id_container.get_ids())
        self.assertEqual([0], IDManager(0).get_function_stack())

        data_packet_0 = DataPacket(ret="Nothing")
        IDManager(0).set_ids_of_packet(data_packet_0)
        self.assertEqual((0, 3), data_packet_0.header.id_container.get_ids())
        self.assertEqual([], IDManager(0).get_function_stack())

    def test_packing(self):
        """
        client        -    server
            func --->
                        <---func
            data --->
                        <---data
        """
        # --------client----------- send func
        func_packet_0 = FunctionPacket("N")
        IDManager(0).set_ids_of_packet(func_packet_0)
        bytes_ = func_packet_0.pack()
        self.assertEqual((0, 0), func_packet_0.header.id_container.get_ids())
        self.assertEqual([0], IDManager(0).get_function_stack())

        # --------server--------- recv func, send func
        byte_stream = ByteStream(bytes_)
        header = Header.from_bytes(byte_stream)
        packet = Packet.from_bytes(header, byte_stream)
        IDManager(1).update_ids_by_packet(packet)
        self.assertEqual((0, 0), packet.header.id_container.get_ids())
        self.assertEqual([0], IDManager(1).get_function_stack())

        func_packet_1 = FunctionPacket("W")
        IDManager(1).set_ids_of_packet(func_packet_1)
        bytes_ = func_packet_1.pack()
        self.assertEqual((1, 1), func_packet_1.header.id_container.get_ids())
        self.assertEqual([0, 1], IDManager(1).get_function_stack())

        # --------client----------- recv func, send data
        byte_stream = ByteStream(bytes_)
        header = Header.from_bytes(byte_stream)
        packet = Packet.from_bytes(header, byte_stream)
        IDManager(0).update_ids_by_packet(packet)
        self.assertEqual((1, 1), packet.header.id_container.get_ids())
        self.assertEqual([0, 1], IDManager(0).get_function_stack())

        data_packet_1 = DataPacket(ret="Nothing")
        IDManager(0).set_ids_of_packet(data_packet_1)
        bytes_ = data_packet_1.pack()
        self.assertEqual((1, 2), data_packet_1.header.id_container.get_ids())
        self.assertEqual([0], IDManager(0).get_function_stack())

        # --------server--------- recv data, send data
        byte_stream = ByteStream(bytes_)
        header = Header.from_bytes(byte_stream)
        packet = Packet.from_bytes(header, byte_stream)
        IDManager(1).update_ids_by_packet(packet)
        self.assertEqual((1, 2), packet.header.id_container.get_ids())
        self.assertEqual([0], IDManager(1).get_function_stack())

        data_packet_0 = DataPacket(ret="Nothing")
        IDManager(1).set_ids_of_packet(data_packet_0)
        bytes_ = data_packet_0.pack()
        self.assertEqual((0, 3), data_packet_0.header.id_container.get_ids())
        self.assertEqual([], IDManager(1).get_function_stack())

        # --------client----------- recv data
        byte_stream = ByteStream(bytes_)
        header = Header.from_bytes(byte_stream)
        packet = Packet.from_bytes(header, byte_stream)
        IDManager(0).update_ids_by_packet(packet)
        self.assertEqual((0, 3), packet.header.id_container.get_ids())
        self.assertEqual([], IDManager(0).get_function_stack())



"""
:module: pynetworking.sides.server
:synopsis: import everything and setup everything for server side code
:author: Julian Sobott
"""
from pynetworking.common import *

global_data.IS_SERVER = True



"""
:module: pynetworking.Packets
:synopsis: Provides classes that are used to send data over the network. The data is packed into packets.
:author: Julian Sobott

Packets:
*********

- `FunctionPacket`: Sends function calls
- `DataPacket`: Sends return messages
- `FileMetaPacket`: Sends file meta data that is needed, when files are transmitted

Every packet class can convert its data to bytes, that can be send over the socket and can convert it back.

**General packet structure:**

- Header:
    - PacketType (Function, Data, FileMeta)
    - ID's
    - specific data size
- Specific Data: (Individual for each packet class

**General packet byte string:**

<function_id><global_id><packet_cls_id><specific_data_len><specific_packet_data>



External_use:
--------------

.. code-block:: python

    packet_one = <cls>Packet(*args, **kwargs) # Look at the individual class for exact signature
    byte_string = packet_one.pack()
    byte_stream = ByteStream(byte_string)
    header = Header.from_bytes(byte_stream)
    packet_two = Packet.from_bytes(header, byte_stream)
    self.assertEqual(packet_one, packet_two)

public classes
-----------------

.. autoclass:: Header
    :members:
    :undoc-members:

.. autoclass:: Packet
    :members:
    :undoc-members:

.. autoclass:: FunctionPacket
    :members:
    :undoc-members:
    :show-inheritance:

.. autoclass:: DataPacket
    :members:
    :undoc-members:
    :show-inheritance:

.. autoclass:: FileMetaPacket
    :members:
    :undoc-members:
    :show-inheritance:

private constants
--------------------

.. autodata:: packets

    Ddict that have for every Packetclass a unique id
"""
from typing import Union, Dict, Any, Callable, Optional

from pynetworking.core.utils import Ddict
from pynetworking.core.Logging import logger
from pynetworking.core.Data import pack_int_type, unpack_int_type, NUM_TYPE_BYTES, \
    general_unpack, general_pack, ByteStream, pack_int


class Header:
    """Every packet has a header. Defines meta data for each packet, that is necessary for network communication.

    :ivar id_container:
    :ivar packet_type:
    :ivar specific_data_size:
    """
    LENGTH_BYTES = 19

    def __init__(self, id_container: 'IDContainer', packet_type: int, specific_data_size: int) -> None:
        self.id_container = id_container
        self.packet_type = packet_type
        self.specific_data_size = specific_data_size

    @classmethod
    def from_packet(cls, packet: 'Packet') -> 'Header':
        from pynetworking.core.ID_management import IDContainer
        packet_type = packets[packet.__class__]
        id_container = IDContainer.default_init()
        specific_data_size = 0
        return cls.__call__(id_container, packet_type, specific_data_size)

    @classmethod
    def from_bytes(cls, byte_stream: ByteStream) -> 'Header':
        from pynetworking.core.ID_management import IDContainer
        id_container = IDContainer.from_bytes(byte_stream)
        packet_type = unpack_int_type(byte_stream.next_bytes(NUM_TYPE_BYTES))
        specific_data_size = byte_stream.next_int()
        return cls.__call__(id_container, packet_type, specific_data_size)

    def pack(self, len_packet_data: int) -> bytes:
        """id's + packet_type + len_packet_data"""
        self.specific_data_size = len_packet_data
        byte_string = b""
        byte_string += self.id_container.pack()
        byte_string += pack_int_type(self.packet_type)
        byte_string += pack_int(len_packet_data)
        return byte_string

    def __eq__(self, other):
        if not isinstance(other, Header):
            return False
        return (self.id_container == other.id_container and
                self.packet_type == other.packet_type and
                self.specific_data_size == other.specific_data_size)

    def __repr__(self):
        return (f"Header({self.id_container.__repr__()}, "
                f"{packets[self.packet_type].__name__}, {self.specific_data_size})\n\t")


class Packet:
    """Abstract super class for all packet classes. A Packet is used to pack and unpack data, to send it over a
    tcp-connection.
    :ivar header:
    """

    def __init__(self, packet: Union['FunctionPacket', 'DataPacket', 'FileMetaPacket']) -> None:
        self.header = Header.from_packet(packet)

    def pack(self) -> bytes:
        raise NotImplementedError()

    def _pack_all(self, specific_data_bytes: bytes) -> bytes:
        """Takes in the packed data from the child-class and creates a fully packed packet from it. Adds header."""
        byte_string = b""
        data_length = len(specific_data_bytes)
        byte_string += self.header.pack(data_length)
        byte_string += specific_data_bytes
        return byte_string

    @classmethod
    def from_bytes(cls, header: Header, byte_stream: ByteStream) -> \
            Union['FunctionPacket', 'DataPacket', 'FileMetaPacket']:
        if header.packet_type not in packets.values():
            raise ValueError("Unknown packet ID: (" + str(header.packet_type) + ")")
        packet = packets[header.packet_type].from_bytes(header, byte_stream)
        packet.header = header
        return packet

    def set_ids(self, function_id: int, outer_id: int) -> None:
        assert isinstance(function_id, int) and isinstance(outer_id, int), f"Ids must be int: ({function_id, outer_id}"
        self.header.id_container.set_ids(function_id, outer_id)

    def __eq__(self, other):
        if isinstance(other, Packet):
            return self.header == other.header
        else:
            return False

    def __repr__(self):
        return str(self.header)


class DataPacket(Packet):
    """Packet to send named data.

    :ivar data: dict with all data."""

    def __init__(self, **kwargs) -> None:
        super().__init__(self)
        self.data = kwargs

    @classmethod
    def from_bytes(cls, header: Header, byte_stream: ByteStream) -> 'DataPacket':
        num_bytes = header.specific_data_size
        data = general_unpack(byte_stream, num_bytes)[0]
        return cls.__call__(**data)

    def pack(self) -> bytes:
        specific_byte_string = general_pack(self.data)
        return super()._pack_all(specific_byte_string)

    def __eq__(self, other):
        if super().__eq__(other) and isinstance(other, DataPacket):
            return self.data == other.data
        else:
            return False

    def __repr__(self):
        string = super().__repr__()
        string += str(self.data)
        return string


class FunctionPacket(Packet):
    """Packet, that stores the function name, and its args. Used to transmit function calls over the network.

    :ivar function_name:
    :ivar args:
    :ivar kwargs:
    """
    def __init__(self, func: Union[Callable, str], *args, **kwargs) -> None:
        super().__init__(self)
        if type(func) is str:
            self.function_name: str = func
        else:
            self.function_name: str = func.__name__
        self.args: tuple = args
        self.kwargs: Dict[str, Any] = kwargs

    @classmethod
    def from_bytes(cls, header: Header, byte_stream: ByteStream) -> 'FunctionPacket':
        num_bytes = header.specific_data_size
        all_data = general_unpack(byte_stream, num_bytes)
        function_name: str = all_data[0]
        args: tuple = all_data[1]
        kwargs: dict = all_data[2]
        return cls.__call__(function_name, *args, **kwargs)

    def pack(self) -> bytes:
        specific_byte_string = general_pack(self.function_name, self.args, self.kwargs)
        return super()._pack_all(specific_byte_string)

    def __eq__(self, other):
        if super().__eq__(other) and isinstance(other, FunctionPacket):
            return self.function_name == other.function_name and self.args == other.args and self.kwargs == other.kwargs
        else:
            return False

    def __repr__(self):
        return f"{super().__repr__()} => FunctionPacket({str(self.function_name)}, " \
            f"{str(self.args)}, {str(self.kwargs)})"


class FileMetaPacket(Packet):
    """Packet that is necessary when files should be sent over the network. Because files may be very big they dont
    want to be packed in one data packet. So to send a file there is the FileMetaClass necessary.

    :ivar src_path: Path where the file is currently located at the sender.
    :ivar dst_path: Path where the file shall be copied to at the receiver.
    :ivar file_size:
    """
    def __init__(self, src_path: str, size: int, dst_path: Optional[str] = None):
        super().__init__(self)
        self.src_path = src_path
        self.dst_path = dst_path
        self.file_size = size

    @classmethod
    def from_bytes(cls, header: Header, byte_stream: ByteStream) -> 'FileMetaPacket':
        num_bytes = header.specific_data_size
        all_data = general_unpack(byte_stream, num_bytes)
        src_path: str = all_data[0]
        dst_path: Optional[str] = all_data[1]
        size: int = all_data[2]
        return cls.__call__(src_path, size, dst_path)

    def pack(self) -> bytes:
        specific_byte_string = general_pack(self.src_path, self.dst_path, self.file_size)
        return super()._pack_all(specific_byte_string)

    def __eq__(self, other):
        if super().__eq__(other) and isinstance(other, FileMetaPacket):
            return self.src_path == other.src_path and self.dst_path == other.dst_path \
                   and self.file_size == other.file_size
        else:
            return False

    def __repr__(self):
        return f"{super().__repr__()} => FileMetaPacket({self.src_path}, {self.file_size}, {str(self.dst_path)})"


packets = Ddict({
    FunctionPacket: 0x101,
    DataPacket: 0x103,
    FileMetaPacket: 0x104
})



"""
:module: pynetworking.Communication_server
:synopsis: Classes that are needed at the server side.
:author: Julian Sobott

public classes
----------------

.. autoclass:: ClientManager
    :members: start, get, mainloop, stop_connections, stop_listening
    :undoc-members:
    :exclude-members: run

.. autoclass:: ClientFunctions
    :members:
    :undoc-members:

private classes
----------------

.. autoclass:: MetaClientManager
    :members:
    :undoc-members:

.. autoclass:: ClientCommunicator
    :members:
    :undoc-members:

private functions
------------------

.. autofunction:: exchange_keys

"""
import threading
import socket
import time
from typing import Dict, Type, Union, Optional

from pynetworking.core.Logging import logger
from pynetworking.core.Communication_general import Communicator, Connector, SocketAddress, Functions, to_server_id
import pynetworking.core.Communication_general
from pynetworking.core.ID_management import IDManager
from pynetworking.core.Packets import DataPacket

__all__ = ["ClientManager", "ClientFunctions"]


class MetaClientManager(type):
    """Allows that multiple ClientManagers can be created. Stores every instance of a ClientManager"""
    _instances: Dict[SocketAddress, 'ClientManager'] = {}
    _last_instance: 'ClientManager' = None

    def __call__(cls, *args, **kwargs) -> 'ClientManager':
        if len(args) == 0:
            if MetaClientManager._last_instance is not None:
                return MetaClientManager._last_instance
            else:
                raise TypeError(
                    "No ClientManager found! First instantiate a ClientManager with address and ClientCommunicator.")
        address: SocketAddress = args[0]
        if address not in cls._instances:
            MetaClientManager._instances[address] = super(MetaClientManager, cls).__call__(*args, **kwargs)
            MetaClientManager._last_instance = cls._instances[address]
        return MetaClientManager._instances[address]

    @staticmethod
    def tear_down():
        while len(MetaClientManager._instances.values()) > 0:
            client_manager: ClientManager = MetaClientManager._instances.popitem()[1]
            client_manager.stop_listening()
            client_manager.stop_connections()


class ClientManager(threading.Thread, metaclass=MetaClientManager):
    """This class accepts new clients and stores them in an array. Provides access to each client. Necessary for every
    server. The :func:`start()` method will start the ClientManager to listen for new clients. It is also possible to
    use this class as context-manager with the `with` statement."""

    def __init__(self, address: SocketAddress = None, client_communicator: Type['ClientCommunicator'] = None,
                 on_client_connect=None, on_client_disconnect=None) -> None:
        super().__init__(name="ClientManager")
        self._socket_connection = socket.socket()
        self._socket_connection.settimeout(1)
        self.clients: Dict[int, ClientCommunicator] = {}
        self._next_client_id = 0
        self._address = address
        self._exit = threading.Event()
        self._client_communicator = client_communicator
        self._on_client_connect = on_client_connect
        self._on_client_disconnect = on_client_disconnect

    def start(self):
        """Start listening to new connections and accepts them"""
        super().start()

    def run(self) -> None:
        try:
            self._socket_connection.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            self._socket_connection.bind(self._address)
            logger.info(f"Server is now listening on: {self._address[0]}:{self._address[1]}")
            self._socket_connection.listen(4)
        except OSError as e:
            # [WinError 10038] socket closed before
            logger.error(e)
            self._exit.set()
        except socket.gaierror:
            raise ValueError(
                f"Address error. {self._address} is not a valid address. Address must be of type {SocketAddress}")

        while not self._exit.is_set():
            try:
                (connection, addr) = self._socket_connection.accept()
                logger.info("New client connected: (%s)", str(addr))
                client_id = self._produce_next_client_id()
                client_communicator_id = to_server_id(client_id)
                client = self._client_communicator(client_communicator_id, self._address, connection,
                                                   self._remove_disconnected_client)
                self._add_client(client_communicator_id, client)
            except OSError as e:
                if not isinstance(e, socket.timeout):
                    if not self._exit.is_set():
                        logger.error("TCP connection closed while listening")
                        # TODO: handle (if possible)

    def _add_client(self, client_communicator_id: int, client: 'ClientCommunicator'):
        self.clients[client_communicator_id] = client
        if self._on_client_connect:
            self._on_client_connect(client)

    def _produce_next_client_id(self) -> int:
        try:
            return self._next_client_id
        finally:
            self._next_client_id += 1

    def get(self, client_id: Optional[int] = None) -> 'ClientCommunicator':
        """Returns the proper ClientCommunicator. The proper one is the one who called the server-side function. This
        function is thread dependent. So if you create a new thread inside the called function you have to store the
        :code:`id` of the current ClientCommunicator and then call this function with this id as optional parameter.
        """
        if client_id is None:
            current_thread: Union[ClientCommunicator, threading.Thread] = threading.current_thread()
            try:
                client_id = current_thread.id
            except AttributeError:
                """Function may only be called from same thread (thread that called the server function) 
                or with a valid existing client_id!"""
                logger.error(
                    "Captain we have a multithreading problem! Thread dependent function called from another thread")
                raise Exception("Thread dependent function called from another thread")
        if client_id not in self.clients.keys():
            raise Exception("No client connected. The client_id doesn't match any connected clients.")
        assert isinstance(self.clients[client_id], ClientCommunicator), "Previous checks didnt handle all cases"
        return self.clients[client_id]

    def mainloop(self):
        """Runs the server till it is stopped by a `KeyboardInterrupt`"""
        while not self._exit.is_set():
            try:
                self._exit.wait(5)
            except KeyboardInterrupt:
                break

    def _remove_disconnected_client(self, communicator: Communicator) -> None:
        """Called when one side stops"""
        try:
            client: ClientCommunicator = self.clients.pop(communicator.get_id())
            if self._on_client_disconnect:
                self._on_client_disconnect(client)
        except KeyError:
            import traceback
            traceback.print_stack()
            logger.error(f"Trying to remove a client that was already removed! {self.clients}: {communicator.get_id()}")

    def stop_listening(self) -> None:
        self._exit.set()
        self._socket_connection.close()
        self.join()
        logger.info("Closed server listener")

    def stop_connections(self) -> None:
        logger.debug(f"Close connections: {self.clients}")
        while len(self.clients.items()) > 0:
            client_id = self.clients.keys().__iter__().__next__()
            client = self.clients[client_id]
            if client.communicator.is_connected():
                client.close_connection()
            else:
                self.clients.pop(client_id)

    def __enter__(self) -> 'ClientManager':
        self.start()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        self.stop_listening()
        self.stop_connections()


class ClientCommunicator(Connector):
    """A static accessible class, that is responsible for communicating with a client.
    This class needs to be overwritten, but is only used internally. The overwritten class needs to set the
    attributes :code:`local_functions` and :code:`remote_functions`.

    :ivar communicator: instance of :class:`pynetworking.Communication_general.Communicator`
    :ivar remote_functions: All functions, that are available at the client side.
        instance of: :class:`pynetworking.Communication_server.ClientFunctions`
    :ivar local_functions: All functions, that are available at the server side.
        instance of: :class:`pynetworking.Communication_client.ServerFunctions`"""

    def __init__(self, id_: int, address: SocketAddress, connection: socket.socket, on_close):
        super().__init__()
        self._id = id_
        self.communicator = Communicator(address, id_, connection, from_accept=True, on_close=on_close,
                                         local_functions=self.local_functions)
        self.communicator.start()
        self.remote_functions.__setattr__(self.remote_functions, "_connector", self)
        if pynetworking.core.Communication_general.ENCRYPTED_COMMUNICATION:
            exchange_keys(self)

    def close_connection(self: Connector, blocking=True, timeout=float("inf")) -> None:
        return super().close_connection(self, blocking, timeout)

    @staticmethod
    def get(client_id: Optional[int] = None, server_address: Optional[SocketAddress] = None) -> 'ClientCommunicator':
        """Same as the :code:`ClientManager.get()`. Not sure which one is more user friendly or intuitive to call"""
        if server_address:
            return ClientManager(server_address).get(client_id)
        else:
            return ClientManager().get(client_id)

    @property
    def id(self):
        return self._id


class ClientFunctions(Functions):
    """Static class that contains all available client side functions. All functions must be stored in the
    :attr:`__dict__` attribute."""
    pass


def exchange_keys(client_communicator: ClientCommunicator):
    """Exchanges a symmetric `communication key` with the client. The `communication key` is encrypted,
    with the public key of the client. After this function, all packets are encrypted with this `communication key`"""
    # generate communication_key#
    cryptographer = client_communicator.communicator.cryptographer
    serialized_communication_key = cryptographer.generate_communication_key()

    # wait for public key
    IDManager(client_communicator.id).append_dummy_functions(2)
    public_key_packet = client_communicator.communicator.wait_for_response()
    serialized_public_key = public_key_packet.data["public_key"]
    cryptographer.public_key_from_serialized_key(serialized_public_key)

    encrypted_communication_key = cryptographer.encrypt_with_public_key(serialized_communication_key)
    # send communication key
    communication_packet = DataPacket(communication_key=encrypted_communication_key)
    client_communicator.communicator.send_packet(communication_packet)
    cryptographer.communication_key = serialized_communication_key

    client_communicator._exchanged_keys = True



"""
:synopsis: A example that shows very brief all important features of the pynetworking module


This is example is meant to demonstrate the use of the most important features of the pynetworking module.
This example has to be adjusted if you want to implement a real client server application. Here client and
server are executed in the same process which is ok for testing but not realistic in reality.

Features:
- call functions with arguments (server->client and client->server)
- get return values
- get Exceptions

What cannot be returned by a function:
- generator
- frame
- traceback
"""
import pynetworking as net

# define the address at which the server listen and the clients connect to
server_address = "127.0.0.1", 5000

# Only log errors
net.Logging.logger.setLevel(40)


def main():
    # open and start the server and close afterwards
    print("This program shows all important features that are possible with the pynetworking library.\n\n")
    with net.ClientManager(server_address, ClientCommunicator):
        # Connect a client to the server
        ServerCommunicator.connect(server_address, blocking=True, timeout=2)

        # create an alias for all server functions (Not necessary, but code looks better)
        server = ServerCommunicator.remote_functions

        # Call functions at the server
        print(f"{'='*7} Calling a function with arguments and return value {'=' * 7}\n")
        types = server.get_types(1, "He", [1, 2], main, lambda a, b: a+b)
        print(types)

        print(f"\n\n{'=' * 7} Calling a function that raises an error {'=' * 7}\n")
        try:
            server.risky_function()
        except NotImplementedError as e:
            print(e)
            print("Error successfully transmitted")

        print(f"\n\n{'=' * 7} Calling a function that calls a function at the client {'=' * 7}\n")
        server.start_communication()

        # Close the connection to the server at client-side
        ServerCommunicator.close_connection()


class ClientFunctions(net.ClientFunctions):
    # A class that defines every method that can be called at a client from the server
    @staticmethod
    def a_client_function(message: str) -> None:
        print(f"At Client: {message}")
        ServerCommunicator.remote_functions.a_communication_function("Hi from Client!")


class ServerFunctions(net.ServerFunctions):
    # A class that defines every method that can be called at the server from a client
    @staticmethod
    def get_types(*args, **kwargs) -> list:
        """returns a list of all types of the passed arguments.
        e.g. simple_function(10, [1,2]) will return [int, list]"""
        arg_types = []
        for arg in args:
            arg_types.append(type(arg))
        for values in kwargs.values():
            arg_types.append(type(values))
        return arg_types

    @staticmethod
    def risky_function():
        raise NotImplementedError("You successfully raised an error!")

    @staticmethod
    def start_communication():
        client: ClientCommunicator = net.ClientManager().get()
        client.remote_functions.a_client_function("Hi from Server!")

    @staticmethod
    def a_communication_function(message):
        print(f"At Server: {message}")


class ClientCommunicator(net.ClientCommunicator):
    # Server-side class that sets the available functions and make them available for communication
    local_functions = ServerFunctions
    remote_functions = ClientFunctions


class ServerCommunicator(net.ServerCommunicator):
    # Client-side class that sets the available functions and make them available for communication
    local_functions = ClientFunctions
    remote_functions = ServerFunctions


if __name__ == '__main__':
    main()



"""
:module: pynetworking.Data
:synopsis: Lowest level module, that manages important data for packets, mostly in bytes.
:author: Julian Sobott

public functions
----------------

.. autofunction:: general_pack

.. autofunction:: general_unpack

.. autofunction:: pack_int_type

.. autofunction:: unpack_int_type

.. autofunction:: pack_int

public classes
--------------

.. autoclass:: ByteStream
    :members:
    :undoc-members:

.. autoclass:: File
    :members:
    :undoc-members:

"""
import os
import dill

from pynetworking.core.utils import Ddict
from pynetworking.core.Logging import logger

NUM_TYPE_BYTES = 3
ENCODING = "utf-8"
BYTEORDER = "big"
NUM_INT_BYTES = 4

types = Ddict({
    int:    0x001,
    float:  0x002,
    str:    0x003,
    list:   0x004,
    dict:   0x005,
    tuple:  0x006,
    bytes:  0x007,
    bool:   0x008,
    type(None):   0x009,
})


def general_pack(*args) -> bytes:
    """Converts all args into a bytes string. """
    return dill.dumps(args)


def general_unpack(byte_stream: 'ByteStream', num_bytes=None) -> tuple:
    """Take in a bytestream, with the bytes string from :func:`general_pack`, and converts it back into a tuple with
    all args."""
    num_bytes = byte_stream.remaining_length if num_bytes is None else num_bytes
    bytes_string = byte_stream.next_bytes(num_bytes)
    data = dill.loads(bytes_string)
    return data


class ByteStream:
    """This class utilises the bytes object. Among other things, it stores the bytes string and the idx. All `next`
    functions move the idx."""

    def __init__(self, byte_string: bytes) -> None:
        self.byte_string = byte_string
        self.idx = 0
        self.length = len(byte_string)
        self.remaining_length = self.length
        self.reached_end = self.remaining_length <= 0

    def next_int(self) -> int:
        """Converts the next bytes into an integer."""
        byte_string = self.next_bytes(NUM_INT_BYTES)
        return int.from_bytes(byte_string, BYTEORDER, signed=True)

    def next_bytes(self, num_bytes: int) -> bytes:
        assert num_bytes >= 0, f"This function is not meant to be called with negative values ({num_bytes})"
        try:
            return self.byte_string[self.idx: self.idx + num_bytes]
        finally:
            self._inc_idx(num_bytes)
            if self.reached_end and self.idx > self.length:
                raise IndexError(f"Byte string ran out of scope: {self.idx} > {self.length}")

    def next_all_bytes(self) -> bytes:
        return self.next_bytes(self.remaining_length)

    def _inc_idx(self, amount: int) -> None:
        self.idx += amount
        self.remaining_length -= amount
        if self.remaining_length <= 0:
            self.reached_end = True

    def remove_consumed_bytes(self) -> None:
        """Deletes all bytes, that are before the idx. Resets all values to fit the new bytes string"""
        self.byte_string = self.byte_string[self.idx:]
        self.length = len(self.byte_string)
        self.remaining_length = self.length
        self.reached_end = self.length == 0
        self.idx = 0

    def __iadd__(self, other: bytes) -> 'ByteStream':
        if not isinstance(other, bytes):
            raise TypeError(f"{type(other)} is not type: bytes")
        self.byte_string += other
        added_length = len(other)
        self.length += added_length
        self.remaining_length += added_length
        self.reached_end = self.remaining_length <= 0
        return self

    def __repr__(self):
        return str(self.byte_string[:self.idx]) + "|" + str(self.byte_string[self.idx:])


class File:
    """This class represents a file that should be sent. If a file is to be sent, an object of this class shall be
    sent with the proper paths. This internally sends the file."""

    def __init__(self, src_path: str, dst_path: str, size=None) -> None:
        self.src_path = src_path
        self.dst_path = dst_path
        if size is None:
            self.size = os.path.getsize(src_path)
        else:
            self.size = size

    @classmethod
    def from_meta_packet(cls, file_meta_packet):
        return cls(file_meta_packet.src_path, file_meta_packet.dst_path, file_meta_packet.file_size)


def pack_int_type(int_type: int) -> bytes:
    """Packs a type described as an integer into bytes"""
    return int.to_bytes(int_type, NUM_TYPE_BYTES, BYTEORDER)


def unpack_int_type(full_byte_string: bytes) -> int:
    """Unpacks bytes into a type described as an integer"""
    return int.from_bytes(full_byte_string[:NUM_TYPE_BYTES], BYTEORDER)


def pack_int(num: int) -> bytes:
    """Packs any integer number into bytes"""
    return int.to_bytes(num, NUM_INT_BYTES, BYTEORDER, signed=True)



"""
:module: pynetworking.sides.client
:synopsis: import everything and setup everything for server side code
:author: Julian Sobott
"""
from pynetworking.common import *

global_data.IS_CLIENT = True



"""
:module: OpenDrive.
:synopsis: 
:author: Julian Sobott

public classes
---------------

.. autoclass:: XXX
    :members:


public functions
----------------

.. autofunction:: XXX

private functions
-----------------


"""
IS_CLIENT = False
IS_SERVER = False



"""
@author: Julian Sobott
@brief:
@description:

@external_use:

@internal_use:

"""
import unittest

from pynetworking.core.Packets import Header, Packet, DataPacket, FunctionPacket, FileMetaPacket
from pynetworking.core.Data import ByteStream
from pynetworking.core.Logging import logger

from pynetworking.tests.example_functions import DummyPerson


class TestPackets(unittest.TestCase):

    def helper_packet_tests(self, packet):
        packet_one = packet
        byte_string = packet_one.pack()
        byte_stream = ByteStream(byte_string)
        header = Header.from_bytes(byte_stream)
        packet_two = Packet.from_bytes(header, byte_stream)
        self.assertEqual(packet_one, packet_two)

    def test_function_packet(self):
        packet = FunctionPacket(example_function, "John", "Miller", (12, 13))
        self.helper_packet_tests(packet)

        packet = FunctionPacket(example_function, "John", tup=("HE",))
        self.helper_packet_tests(packet)

    def test_function_packet_class_arg(self):
        packet = FunctionPacket(example_function, DummyPerson("He", 12))
        self.helper_packet_tests(packet)

    def test_data_packet(self):
        packet = DataPacket(Name="John Miller")
        self.helper_packet_tests(packet)

        packet = DataPacket(username="John", age="28", password={"val": ["he", "he"], "val2": (2, (3, 5))})
        self.helper_packet_tests(packet)

        packet = DataPacket(classObj=TestHeader)
        self.helper_packet_tests(packet)

    def test_file_meta_packet(self):
        packet = FileMetaPacket(r"C:\Hello\World\src.txt", 0, "D:/WOW/dst.txt")
        self.helper_packet_tests(packet)

        packet = FileMetaPacket(r"C:\Hello\World\src.txt", 0)
        self.helper_packet_tests(packet)


class TestHeader(unittest.TestCase):

    def test_packing(self):
        packet = DataPacket(Test=10)
        header1_0 = packet.header
        byte_string = packet.pack()
        byte_stream = ByteStream(byte_string)
        header1_1 = Header.from_bytes(byte_stream)
        self.assertEqual(header1_0, header1_1)


def example_function(name, second_name="Miller", tup=()):
    print("Hello: " + str(name) + " " + str(second_name) + " --> " + str(tup))



import unittest
import sys
import os
import time

from thread_testing import get_num_non_dummy_threads, wait_till_joined, wait_till_condition

from pynetworking.core.Communication_client import ServerCommunicator, MultiServerCommunicator, ServerFunctions
from pynetworking.core.Communication_server import ClientManager, ClientFunctions, ClientCommunicator, MetaClientManager
from pynetworking.core.Communication_general import to_server_id
import pynetworking.core.Communication_general
from pynetworking.core.Logging import logger

from pynetworking.tests.example_functions import DummyPerson, DummyServerCommunicator, DummyClientCommunicator, \
    DummyMultiServerCommunicator

dummy_address = ("localhost", 5000)
server_address = ("0.0.0.0", 5000)

logger.setLevel(10)


class CommunicationTestCase(unittest.TestCase):
    def tearDown(self):
        DummyServerCommunicator.close_connection()
        MultiServerCommunicator.close_all_connections()
        MetaClientManager.tear_down()

    def setUp(self):
        DummyMultiServerCommunicator.close_all_connections()
        DummyServerCommunicator.close_connection()


class StdOutEqualizer:

    def __init__(self, test_case, expected):
        self.test_case: CommunicationTestCase = test_case
        self.expected = expected
        self.original = sys.stdout
        self.file_path = os.path.join(os.path.split(__file__)[0], "std_out.txt")

    def __enter__(self):
        self.original = sys.stdout
        sys.stdout = open(self.file_path, "w")
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        sys.stdout.close()
        sys.stdout = self.original
        with open(self.file_path, "r") as std_out:
            self.test_case.assertEqual(self.expected, std_out.read().strip())


class TestConnecting(CommunicationTestCase):

    def setUp(self):
        DummyServerCommunicator._exchanged_keys = False

    def test_single_client_connect(self):
        DummyServerCommunicator.connect(dummy_address, timeout=0)
        self.assertEqual(DummyServerCommunicator.remote_functions.__getattr__("_connector"), DummyServerCommunicator)
        self.assertEqual(DummyServerCommunicator.remote_functions().__getattr__("_connector"), DummyServerCommunicator)
        self.assertEqual(DummyServerCommunicator.remote_functions, DummyServerCommunicator.remote_functions())

    def test_multi_client_connect(self):
        DummyMultiServerCommunicator(0).connect(dummy_address, timeout=0)
        self.assertIsInstance(DummyMultiServerCommunicator(0).remote_functions.__getattr__("_connector"),
                              DummyMultiServerCommunicator)
        DummyMultiServerCommunicator(1).connect(dummy_address, timeout=0)
        self.assertIsInstance(DummyMultiServerCommunicator(1).remote_functions.__getattr__("_connector"),
                              DummyMultiServerCommunicator)

    def test_single_client(self):
        with ClientManager(server_address, DummyClientCommunicator) as listener:
            wait_till_condition(lambda : False is True, timeout=2)
            DummyServerCommunicator.connect(dummy_address)

            wait_till_condition(lambda: len(listener.clients) == 1, timeout=1)

            self.assertEqual(len(listener.clients), 1)
            self.assertEqual(get_num_non_dummy_threads(), 4)  # Main, listener, client_communicator, server_communicator
            self.assertEqual(listener.clients[to_server_id(0)].communicator._socket_connection.getpeername(),
                             DummyServerCommunicator.communicator._socket_connection.getsockname())
            DummyServerCommunicator.close_connection()
            DummyServerCommunicator.close_connection()
            wait_till_joined(DummyServerCommunicator.communicator, timeout=1)
            wait_till_condition(lambda: len(listener.clients.values()) == 0, timeout=1)
            # wait_till_condition(lambda: get_num_non_dummy_threads() == 2)

            self.assertEqual(get_num_non_dummy_threads(), 2)  # Main, listener
        wait_till_joined(listener, timeout=1)

        self.assertEqual(get_num_non_dummy_threads(), 1)

    def test_multiple_clients(self):
        pynetworking.core.Communication_general.set_encrypted_communication(False)
        with ClientManager(server_address, DummyClientCommunicator) as listener:
            DummyMultiServerCommunicator(0).connect(dummy_address)
            DummyMultiServerCommunicator(1).connect(dummy_address)

            wait_till_condition(lambda: len(listener.clients) == 2, timeout=2)

            self.assertEqual(len(listener.clients), 2)

            self.assertEqual(get_num_non_dummy_threads(), 6)
            # Main, listener, 2 * client_communicator, 2 * server_communicator
            self.assertEqual(listener.clients[to_server_id(0)].communicator._socket_connection.getpeername(),
                             DummyMultiServerCommunicator(0).communicator._socket_connection.getsockname())
            self.assertEqual(listener.clients[to_server_id(1)].communicator._socket_connection.getpeername(),
                             DummyMultiServerCommunicator(1).communicator._socket_connection.getsockname())

            DummyMultiServerCommunicator(0).close_connection()
            DummyMultiServerCommunicator(1).close_connection()

            wait_till_joined(DummyMultiServerCommunicator(0).communicator, timeout=2)
            wait_till_joined(DummyMultiServerCommunicator(1).communicator, timeout=2)
            wait_till_condition(lambda: len(listener.clients.values()) == 0, timeout=2)
            self.assertEqual(get_num_non_dummy_threads(), 2)  # Main, listener
        wait_till_joined(listener, timeout=1)

        self.assertEqual(get_num_non_dummy_threads(), 1)
        pynetworking.core.Communication_general.set_encrypted_communication(True)

    def test_offline_server(self):
        i = 0
        while not DummyServerCommunicator.connect(dummy_address, timeout=2):
            i += 1
            if i == 3:
                break
        self.assertEqual(get_num_non_dummy_threads(), 1)
        self.assertEqual(i, 3)

    def test_server_turn_on(self):
        DummyServerCommunicator.connect(dummy_address, blocking=False)
        with ClientManager(server_address, DummyClientCommunicator) as listener:
            wait_till_condition(lambda: DummyServerCommunicator._exchanged_keys is True, timeout=4)
            self.assertEqual(len(listener.clients), 1)
            self.assertEqual(True, DummyServerCommunicator._exchanged_keys)
            DummyServerCommunicator.close_connection()
            wait_till_joined(DummyServerCommunicator.communicator, timeout=2)
            wait_till_condition(lambda: len(listener.clients) == 0, timeout=2)
            self.assertEqual(len(listener.clients), 0)
        self.assertEqual(get_num_non_dummy_threads(), 1)

    def test_listener_clients(self):
        with ClientManager(server_address, DummyClientCommunicator) as listener:
            self.assertEqual(len(listener.clients.values()), 0)
            DummyServerCommunicator.connect(dummy_address)
            wait_till_condition(lambda: len(listener.clients) == 1, timeout=2)
            self.assertEqual(len(listener.clients.values()), 1)
            DummyServerCommunicator.close_connection()
            wait_till_condition(lambda: len(listener.clients) == 0, timeout=2)
            self.assertEqual(len(listener.clients.values()), 0)


class TestCommunicating(CommunicationTestCase):

    def helper_test_func(self, func, args: tuple, expected_ret):
        with ClientManager(server_address, DummyClientCommunicator):
            DummyServerCommunicator.connect(dummy_address)
            try:
                ret_value = func(*args)
                self.assertEqual(expected_ret, ret_value)
            except TimeoutError:
                self.fail("TimeoutError")

    # Error cases
    def test_functions_no_connection(self):
        self.assertRaises(ConnectionError, DummyServerCommunicator.remote_functions(timeout=0).dummy_no_arg_no_ret)

    def test_functions_timeout(self):
        # May fail if packets are handled at server is implemented
        with ClientManager(dummy_address, DummyClientCommunicator):
            DummyServerCommunicator.connect(dummy_address)
            self.assertRaises(TimeoutError, DummyServerCommunicator.remote_functions(timeout=0).dummy_no_arg_no_ret)

    # Succeed cases
    def test_no_arg_no_ret(self):
        with StdOutEqualizer(self, "no_arg_no_ret() called"):
            self.helper_test_func(DummyServerCommunicator.remote_functions(timeout=2).no_arg_no_ret,
                                  tuple(),
                                  None)

    def test_no_arg_ret(self):
        with StdOutEqualizer(self, "no_arg_ret() called"):
            self.helper_test_func(DummyServerCommunicator.remote_functions(timeout=2).no_arg_ret,
                                  tuple(),
                                  True)

    def test_immutable_args_ret(self):
        name = "Anne"
        age = 78
        children = ("Tom", "Pia")
        self.helper_test_func(DummyServerCommunicator.remote_functions(timeout=2).immutable_args_ret,
                              (name, age, children),
                              f"{name} is {age} old and has {len(children)}: {children}")

    def test_args_ret_object(self):
        name = "Anne"
        age = 78
        self.helper_test_func(DummyServerCommunicator.remote_functions(timeout=2).args_ret_object,
                              (name, age),
                              DummyPerson(name, age))

    def test_class_args_ret(self):
        name = "Anne"
        age = 78
        self.helper_test_func(DummyServerCommunicator.remote_functions(timeout=2).class_args_ret,
                              (DummyPerson(name, age),),
                              (name, age))

    def test_huge_args_huge_ret(self):
        arg1 = "Text" * 1200
        arg2 = DummyPerson("John", -9100)
        arg3 = "End"
        args = (arg1, arg2, arg3)
        self.helper_test_func(DummyServerCommunicator.remote_functions(timeout=2).huge_args_huge_ret,
                              args,
                              args)

    def test_small_file(self):
        import os
        file_path = os.path.join(os.path.split(__file__)[0], "std_out.txt")
        destination_path = os.path.join(os.path.split(__file__)[0], "new_file.txt")
        with ClientManager(server_address, DummyClientCommunicator):
            DummyServerCommunicator.connect(dummy_address)
            file = DummyServerCommunicator.remote_functions().get_file(file_path, destination_path)
            self.assertEqual(file.src_path, file_path)
            self.assertEqual(file.dst_path, destination_path)
        self.assertEqual(os.path.getsize(file_path), os.path.getsize(destination_path))

    def test_huge_file(self):
        import os
        file_path = os.path.join(os.path.split(__file__)[0], "dummy.txt")
        with open(file_path, "w+") as f:
            for i in range(400):
                f.write("H"*100000)

        destination_path = os.path.join(os.path.split(__file__)[0], "new_file.txt")
        with ClientManager(server_address, DummyClientCommunicator):
            DummyServerCommunicator.connect(dummy_address)
            file = DummyServerCommunicator.remote_functions().get_file(file_path, destination_path)
            self.assertEqual(file.src_path, file_path)
            self.assertEqual(file.dst_path, destination_path)
        self.assertEqual(os.path.getsize(file_path), os.path.getsize(destination_path))

    def test_func_in_func(self):
        self.helper_test_func(DummyServerCommunicator.remote_functions(timeout=2).func_in_func,
                              (0,),
                              2)

    def test_many_func_in_func(self):
        self.helper_test_func(DummyServerCommunicator.remote_functions(timeout=2).many_func_in_func,
                              tuple(),
                              120)


if __name__ == '__main__':
    unittest.main()



"""
:synopsis: All functions located at the server_side
"""
import pynetworking as net


def greet_client(name: str):
    print(f"Hello {name}")
    ret_value = net.ClientManager().get().remote_functions.client_func("Paul")
    if ret_value:
        print("Cool you responded")
    else:
        print("Mhhh something went wrong ")
    return "Goodbye"


def server_faculty(number: int) -> int:
    if number <= 1:
        return number
    return number * net.ClientManager().get().remote_functions.client_faculty(number - 1)





"""
:module: pynetworking.Communication_server
:synopsis: Classes that are needed at the client side.
:author: Julian Sobott

public classes
----------------

.. autoclass:: ServerCommunicator
    :members:
    :undoc-members: connect

.. autoclass:: MultiServerCommunicator
    :members:
    :undoc-members: connect

.. autoclass:: ServerFunctions
    :members:
    :undoc-members:


private functions
------------------

.. autofunction:: exchange_keys

"""
from typing import Union, Type, Optional

from pynetworking.core.Logging import logger
from pynetworking.core.Communication_general import Connector, SingleConnector, MultiConnector, Functions, SocketAddress
from pynetworking.core.Packets import DataPacket
from pynetworking.core.ID_management import IDManager


class ServerCommunicator(SingleConnector):
    """A static accessible class, that is responsible for communicating with the server.
        This class needs to be overwritten. The overwritten class needs to set the attributes :code:`local_functions` and
        :code:`remote_functions`. To call a function at the server type:
        :code:`ServerCommunicator.remote_functions.dummy_function(x, y)`
        """

    @classmethod
    def connect(cls, addr: SocketAddress, blocking=True, timeout=float("inf"), **kwargs) -> Optional[bool]:
        connected = super().connect(addr, blocking, timeout, exchange_keys)
        return connected


class MultiServerCommunicator(MultiConnector):
    """A class that allows in contrast to the :class:`ServerCommunicator` multiple instances. This class also needs
    to be overwritten, just like :class:`ServerCommunicator`. To create and use call :code:`MultiServerCommunicator(n)`,
    where n is any number below 30. The object may not be stored, but can be called in different parts of the Code and
    the same object is returned, like a Singleton.
    """

    def connect(self, addr: SocketAddress, blocking=True, timeout=float("inf"), **kwargs) -> Optional[bool]:
        connected = super().connect(addr, blocking, timeout, exchange_keys)
        return connected


class ServerFunctions(Functions):
    """Static class that contains all available server side functions. All functions must be stored in the
    :attr:`__dict__` attribute."""
    pass


def exchange_keys(connector: Union['Connector', Type['SingleConnector']]):
    """Exchanges a symmetric `communication key` with the server. The `communication key` is received from the
    server. It is decrypted with the private key. After this function, all packets are encrypted with this
    `communication key`"""
    # generate public key + private key
    cryptographer = connector.communicator.cryptographer
    cryptographer.generate_key_pair()
    serialized_public_key = cryptographer.get_serialized_public_key()
    IDManager(connector.get_id()).append_dummy_functions(2)
    # send public key
    public_key_packet = DataPacket(public_key=serialized_public_key)
    connector.communicator.send_packet(public_key_packet)
    # wait for communication key
    communication_packet = connector.communicator.wait_for_response()
    encrypted_communication_key = communication_packet.data["communication_key"]
    # decrypt key with private key
    communication_key = cryptographer.decrypt_with_private_key(encrypted_communication_key)
    # set communication key
    cryptographer.communication_key = communication_key
    connector._exchanged_keys = True



"""

"""
import time
import unittest
import threading
from pynetworking import ClientManager, MultiServerCommunicator
from pynetworking.tests.example_functions import DummyServerCommunicator, DummyMultiServerCommunicator, \
    DummyClientCommunicator
from pynetworking.core.Logging import logger
from pynetworking.core import Communication_general

logger.setLevel(10)


class TestManyClients(unittest.TestCase):
    address = "127.0.0.1", 5000

    def setUp(self):
        self.manager = ClientManager(self.address, DummyClientCommunicator)
        self.manager.start()
        Communication_general.ENCRYPTED_COMMUNICATION = False

    def tearDown(self):
        DummyMultiServerCommunicator.close_all_connections()
        logger.debug(self.manager.clients)
        self.manager.stop_listening()
        self.manager.stop_connections()

    def test_many(self):
        clients = []
        thread_pool = []

        def client_exe(client: DummyMultiServerCommunicator):
            client.connect(self.address)
            c_id = client.remote_functions.return_client_id()
            print(c_id)
            time.sleep(5)
            client.close_connection()

        for i in range(20):
            clients.append(DummyMultiServerCommunicator(i))
            thread_pool.append(threading.Thread(target=client_exe, args=(clients[i],)))
            thread_pool[i].start()
            time.sleep(0.1)

        time.sleep(1)
        logger.debug("Finished")


from setuptools import setup, find_packages
import unittest

with open("README.rst", "r") as fh:
    long_description = fh.read()


def my_test_suite():
    from pynetworking.Logging import logger
    logger.setLevel(0)
    test_loader = unittest.TestLoader()
    test_suite = test_loader.discover('pynetworking.tests', pattern='test_*.py')
    return test_suite


def main(version: str = "0.1.1"):
    import sys
    print(sys.argv)
    setup(name='pynetworking',
          version=version,
          description='High level network communication',
          long_description=long_description,
          url='https://github.com/JulianSobott/pynetworking',
          author='Julian Sobott',
          author_email='julian.sobott@gmx.de',
          license='Apache',
          packages=find_packages(),
          test_suite='setup.my_test_suite',
          include_package_data=True,
          keywords='network packet communication',
          project_urls={
              "Bug Tracker": "https://github.com/JulianSobott/pynetworking/issues",
              "Documentation": "http://pynetworking.readthedocs.io/",
              "Source Code": "https://github.com/JulianSobott/pynetworking",
          },
          classifiers=[
              "Programming Language :: Python :: 3.7",
              "License :: OSI Approved :: Apache Software License",
              "Operating System :: OS Independent",
              "Topic :: Software Development :: Libraries",
          ],
          zip_safe=False,
          install_requires=['cryptography', 'dill']
          )


if __name__ == '__main__':
    main()



"""
:module: pynetworking.ID_management
:author: Julian Sobott
:created: 13.11.2018
:synopsis: Handles ids for proper network communication

Each packet has 2 ids: function_id - global_id

- function_id: increased for each new function. FunctionPacket and DataPacket have the same function_id

        Which data_packet belongs to which function

- global_id: increased for each new packet.

        See which packet is next. Watch packet loss

External use
-------------

Call :class:`IDManager(ID)` to work with it. ID is the ID of the communicator (per client one ID). You don't need
to store an object of :class:`IDManager`, because this is handled by its metaclass. When a communicator is closed
call :func:`remove_manger`, with the communicator ID.

public functions
----------------

.. autofunction:: remove_manager

public classes
--------------

.. autoclass:: IDManager
    :members:
    :undoc-members:


.. autoclass:: IDContainer
    :members:
    :undoc-members:

private classes
----------------

.. autoclass:: MetaIDManager
    :members:

"""
from typing import List, Optional, Dict, Tuple

from pynetworking.core.Data import NUM_INT_BYTES, NUM_TYPE_BYTES, pack_int
from pynetworking.core.Logging import logger
from pynetworking.core.Packets import FunctionPacket, DataPacket, FileMetaPacket, Packet

__all__ = ["IDManager", "remove_manager", "IDContainer"]


class MetaIDManager(type):
    """MetaIDManager is a metaclass to deliver multiple IDManagers for multiple Communicators. Each Communicator has
    its own ID. With this ID it gets its appropriate IDManager"""
    _instances: Dict[int, 'IDManager'] = {}

    def __call__(cls, *args, **kwargs):
        id_ = args[0]
        if id_ not in cls._instances:
            MetaIDManager._instances[id_] = super(MetaIDManager, cls).__call__(id_)
        return MetaIDManager._instances[id_]

    @staticmethod
    def remove(id_: int):
        MetaIDManager._instances.pop(id_)


class IDManager(metaclass=MetaIDManager):
    """Sets IDs for each packet."""

    def __init__(self, id_: int) -> None:
        self.id = id_
        self._next_function_id = 0
        self._next_global_id = 0
        self._function_stack: List[int] = []

    def set_ids_of_packet(self, packet: Packet) -> Optional[Packet]:
        """Set ids of packet and adjust internal state"""
        global_id = self._next_global_id
        if isinstance(packet, FunctionPacket):
            func_id = self._is_function_packet()
        elif isinstance(packet, DataPacket) or isinstance(packet, FileMetaPacket):
            func_id = self._is_data_packet()
        else:
            logger.error("Unknown packet_class (%s)", type(packet).__name__)
            return None

        packet.set_ids(func_id, global_id)
        self._next_global_id += 1
        return packet

    def update_ids_by_packet(self, packet: Packet) -> None:
        """Called every time a new packet arrives."""
        self._next_global_id = packet.header.id_container.global_id + 1
        if isinstance(packet, FunctionPacket):
            self._is_function_packet()
        elif isinstance(packet, DataPacket) or isinstance(packet, FileMetaPacket):
            self._is_data_packet()
        else:
            logger.error("Unknown packet_class (%s)", type(packet).__name__)

    def get_next_outer_id(self) -> int:
        return self._next_global_id

    def _is_function_packet(self) -> int:
        self._function_stack.append(self._next_function_id)
        function_id = self._next_function_id

        self._next_function_id += 1
        return function_id

    def _is_data_packet(self) -> int:
        try:
            function_id = self._function_stack.pop()
            return function_id
        except IndexError:
            logger.error("Trying to pop a empty function stack")
            return -1

    def get_next_ids(self) -> Tuple[int, int]:
        return self._next_function_id, self._next_global_id

    def get_function_stack(self) -> List[int]:
        return self._function_stack

    def append_dummy_functions(self, num=1):
        for i in range(num):
            self._function_stack.append(-2)

    def __repr__(self):
        return f"IDManager_{self.id}({self._next_function_id, self._next_global_id})"


def remove_manager(id_: int) -> None:
    """Called when a communicator is stopped. So its ID Manager isn´t needed anymore"""
    try:
        MetaIDManager.remove(id_)
    except KeyError:
        pass


class IDContainer:
    """Stores the id`s of a packet. Packs and unpacks itself at the send process.

    Every packet has 2 id`s. A :attr:`function_id`. This one stores the function it belongs to. This way return
    packets can be unambiguously matched to the proper function. The second id is the :attr:`global_id`. It is
    incremented for each packet. This way packet loss can be detected and packets are handled in the correct order.
    """
    TOTAL_BYTE_LENGTH = 3 * NUM_INT_BYTES + 3 * NUM_TYPE_BYTES

    def __init__(self, function_id: int, outer_id: int) -> None:
        self.function_id = function_id
        self.global_id = outer_id

    @classmethod
    def default_init(cls):
        """Set id`s that must be changed at sending."""
        return cls.__call__(-1, -1)

    def pack(self) -> bytes:
        byte_string = pack_int(self.function_id)
        byte_string += pack_int(self.global_id)
        return byte_string

    @classmethod
    def from_bytes(cls, byte_stream: 'ByteStream') -> 'IDContainer':
        function_id = byte_stream.next_int()
        outer_id = byte_stream.next_int()
        return cls.__call__(function_id, outer_id)

    def set_ids(self, function_id: int, outer_id: int):
        self.function_id = function_id
        self.global_id = outer_id

    def get_ids(self) -> Tuple[int, int]:
        return self.function_id, self.global_id

    def __repr__(self):
        return f"IDContainer({str(self.function_id)}, {str(self.global_id)})"

    def __eq__(self, other):
        if not isinstance(other, IDContainer):
            return False
        return (self.function_id == other.function_id and
                self.global_id == other.global_id)



"""
@author: Julian Sobott
@brief:
@description:

@external_use:

@internal_use:
"""
from unittest import TestCase

from pynetworking.core.Data import general_pack, general_unpack, ByteStream

from pynetworking.tests.example_functions import DummyPerson


def single_value(test_self, value):
    byte_string = general_pack(value)
    byte_stream = ByteStream(byte_string)
    new_value = general_unpack(byte_stream)[0]
    test_self.assertEqual(value, new_value)


class TestPacking(TestCase):

    def test_int(self):
        single_value(self, -890)

    def test_float(self):
        single_value(self, 212400.7598823849019)

    def test_str_special_chars(self):
        string = "!$Hqä \n}\t."
        single_value(self, string)

    def test_str_long(self):
        # Warning: strings with length more than 2**30 will crash the PC
        string = "W" * 2**20
        single_value(self, string)

    def test_list(self):
        value = [[i for i in range(2)]].append([i for i in range(2)])
        single_value(self, value)

    def test_dict(self):
        value = {"val1": 1, "väl2": "Hä llo", "val3": [((2, 6), 1), 7]}
        single_value(self, value)

    def test_tuple(self):
        single_value(self, ("Hello", 10, [2, 0]))
        single_value(self, ())

    def test_bool(self):
        single_value(self, True)

    def test_bytes(self):
        single_value(self, b"Hello")

    def test_None(self):
        single_value(self, None)

    def test_object(self):
        single_value(self, DummyPerson("John", 90))

    def test_lambda(self):
        byte_string = general_pack(lambda x, y: x + y)
        byte_stream = ByteStream(byte_string)
        new_value = general_unpack(byte_stream)[0]
        self.assertEqual(10, new_value(5, 5))



